#include <math.h>
#include <stdlib.h>
#include "../epsilon.h"
#include "../linear_algebra.h"
#include "../linear_programming.h"
#include "../strawberry_fields.h"

enum {
  SF_TAX = 1
};

static void pivot(int e, int l, struct lp_basis *basis) {
  int m = basis->m;
  double *a = basis->a;
  for (int i = 0; i < m; i++) {
    double a_i_e = a[e * m + i];
    a[e * m + i] = a[i * m + l];
    a[i * m + l] = a_i_e;
  }
  double *c = basis->c;
  double c_e = c[e];
  c[e] = c[l];
  c[l] = c_e;
  int *q = basis->q;
  int q_e = q[e];
  q[e] = q[l];
  q[l] = q_e;
}

void sf_fast_simplex(const struct sf_input *input,
                     struct sf_interval2 box,
                     double upper_bound,
                     struct sf_node *node) {
  double eps = epsilon();
  struct sf_point_list s_list;
  sf_strawberry_list_from_input(input, &s_list);
  struct sf_point_list rep_list;
  struct lp_basis dual;
  while (1) {
    struct sf_equivalence_relation relation;
    sf_label_equivalence_classes(box, node->solution, &relation);
    sf_choose_representatives(box, &relation, &s_list, &rep_list);
    int m = node->solution.n_greenhouses, n = rep_list.n_points;
    lp_allocate(m, n + SF_TAX, &dual);
    lp_initialize(&dual);
    for (int j = 0; j < n; j++) {
      for (int i = 0; i < m; i++) {
        dual.a[(m + j) * m + i] = sf_point_in_interval2(rep_list.point[j], node->solution.greenhouse[i]);
      }
    }
    for (int i = 0; i < m; i++) {
      dual.b[i] = sf_cost(node->solution.greenhouse[i]);
    }
    for (int j = 0; j < n; j++) {
      dual.c[m + j] = 1.0;
    }
    if (SF_TAX) {
      for (int i = 0; i < m; i++) {
        dual.a[(m + n) * m + i] = -1.0;
      }
      dual.c[m + n] = -input->n_greenhouses_max;
    }
    int e = lp_maximize(&dual);
    for (int k = 0; k < rep_list.n_points; k++) {
      struct sf_point p = rep_list.point[k];
      struct sf_label *label = relation.label[p.i][p.j];
      label->y = dual.qix[m + k] / (label->n_non_reps + 1);
    }
    node->tax = SF_TAX ? dual.qix[m + n] : 0.0;
    for (int k = 0; k < s_list.n_points; k++) {
      struct sf_point p = s_list.point[k];
      node->y[p.i][p.j] = relation.label[p.i][p.j]->y;
    }
    sf_sums_from_prices(box, node->y, node->y_sum);
    if (e == -1) {
      break;
    }
    lp_free(&dual);
    struct sf_interval2 g_mvc;
    double g_mvc_badness = eps;
    for (int k = 0; k < node->list.n_greenhouses; k++) {
      struct sf_interval2 g = node->list.greenhouse[k];
      double g_badness = sf_sum(node->y_sum, g) - node->tax;
      if (g_badness > g_mvc_badness) {
        g_mvc = g;
        g_mvc_badness = g_badness;
      }
    }
    if (!(g_mvc_badness > eps)) {
      node->super.key = INFINITY;
      return;
    }
    sf_append(g_mvc, &node->solution);
  }
  node->super.key = la_dot_product(dual.b, dual.y, dual.m);
  struct lp_basis primal;
  {
    int m = dual.m, n = dual.n;
    lp_allocate(n, m, &primal);
    for (int k = 0; k < m + n; k++) {
      int q_k = dual.q[k];
      primal.q[k < m ? n + k : k - m] = q_k < m ? n + q_k : q_k - m;
    }
    for (int k = 0; k < m; k++) {
      primal.qix[n + k] = dual.y[k];
    }
    lp_free(&dual);
  }
  while (1) {
    struct sf_interval2 g_mvc;
    double g_mvc_badness = eps;
    for (int k = 0; k < node->list.n_greenhouses; k++) {
      struct sf_interval2 g = node->list.greenhouse[k];
      double g_badness = sf_sum(node->y_sum, g) - node->tax - sf_cost(g);
      if (g_badness > g_mvc_badness) {
        g_mvc = g;
        g_mvc_badness = g_badness;
      }
    }
    if (!(g_mvc_badness > eps)) {
      int n_active = 0;
      for (int k = 0; k < node->solution.n_greenhouses; k++) {
        if (primal.qix[primal.m + k] > eps) {
          node->solution.greenhouse[n_active++] = node->solution.greenhouse[k];
        }
      }
      node->solution.n_greenhouses = n_active;
      break;
    }
    sf_append(g_mvc, &node->solution);

    int *active = calloc(primal.m + primal.n, sizeof *active);
    for (int k = 0; k < primal.m; k++) {
      active[primal.q[k]] = 1;
    }
    const int *active_greenhouse = active + primal.m;
    lp_free(&primal);
    struct sf_point_list inactive_rep_list;
    inactive_rep_list.n_points = 0;
    for (int k = 0; k < rep_list.n_points; k++) {
      if (!active[k]) {
        inactive_rep_list.point[inactive_rep_list.n_points++] = rep_list.point[k];
      }
    }
    struct sf_equivalence_relation relation;
    sf_label_equivalence_classes(box, node->solution, &relation);
    sf_choose_representatives(box, &relation, &s_list, &rep_list);
    for (int k = 0; k < inactive_rep_list.n_points; k++) {
      struct sf_point p = inactive_rep_list.point[k];
      relation.label[p.i][p.j]->mark = -3;
    }
    int n_active_reps = rep_list.n_points;
    for (int k = 0; k < n_active_reps; ) {
      struct sf_point p = rep_list.point[k];
      if (relation.label[p.i][p.j]->mark == -3) {
        rep_list.point[k] = rep_list.point[--n_active_reps];
        rep_list.point[n_active_reps] = p;
      } else {
        k++;
      }
    }
    int m = rep_list.n_points + SF_TAX;
    lp_allocate(m, node->solution.n_greenhouses, &primal);
    lp_initialize(&primal);
    for (int j = 0; j < node->solution.n_greenhouses; j++) {
      for (int i = 0; i < rep_list.n_points; i++) {
        primal.a[(m + j) * m + i] = -sf_point_in_interval2(rep_list.point[i], node->solution.greenhouse[j]);
      }
    }
    for (int i = 0; i < rep_list.n_points; i++) {
      primal.b[i] = -1.0;
    }
    for (int j = 0; j < node->solution.n_greenhouses; j++) {
      primal.c[m + j] = -sf_cost(node->solution.greenhouse[j]);
    }
    if (SF_TAX) {
      for (int j = 0; j < node->solution.n_greenhouses; j++) {
        primal.a[(m + j) * m + rep_list.n_points] = 1.0;
      }
      primal.b[rep_list.n_points] = input->n_greenhouses_max;
    }
    for (int j = 0; j < node->solution.n_greenhouses - 1; j++) {
      if (active_greenhouse[j]) {
        pivot(m + j, n_active_reps++, &primal);
      }
    }
    free(active);
    lp_maximize(&primal);
    node->super.key = -la_dot_product(primal.c, primal.x, primal.m);
    for (int k = 0; k < rep_list.n_points; k++) {
      struct sf_point p = rep_list.point[k];
      struct sf_label *label = relation.label[p.i][p.j];
      label->y = primal.y[k] / (label->n_non_reps + 1);
    }
    node->tax = SF_TAX ? primal.y[rep_list.n_points] : 0.0;
    for (int k = 0; k < s_list.n_points; k++) {
      struct sf_point p = s_list.point[k];
      node->y[p.i][p.j] = relation.label[p.i][p.j]->y;
    }
    sf_sums_from_prices(box, node->y, node->y_sum);
  }
  lp_free(&primal);
  node->super.key = ceil((1.0 - eps) * node->super.key);
}
