#include <math.h>
#include <stdlib.h>
#include "../strawberry_fields.h"

enum {
  SF_DEBUG = 0
};

static struct sf_node *allocate(struct sf_node **last, struct sf_node *parent, int *next_id) {
  struct sf_node *node = malloc(sizeof *node);
  node->previous = *last;
  *last = node;
  node->parent = parent;
  node->active = NULL;
  node->id = (*next_id)++;
  return node;
}

static struct pq_node *evaluate_node(const struct sf_input *input,
                                     struct sf_interval2 box,
                                     struct pq_node *queue,
                                     struct sf_node *node,
                                     struct sf_greenhouse_list *solution,
                                     double *upper_bound) {
  sf_activate(*upper_bound, node);
  sf_fast_simplex(input, box, *upper_bound, node);
  if (SF_DEBUG) {
    fprintf(stderr, "%d: super.key = %g", node->id, node->super.key);
    if (node->parent) {
      fprintf(stderr, "; parent = %d", node->parent->id);
    }
  }
  if (node->super.key < *upper_bound) {
    if (sf_select_branch(box, node->solution, node)) {
      queue = pq_push(queue, (struct pq_node *)node);
      if (SF_DEBUG) {
        fprintf(stderr, "; p = %d,%d; q = %d,%d", node->p.i, node->p.j, node->q.i, node->q.j);
      }
    } else {
      free(solution->greenhouse);
      *solution = sf_copy(node->solution);
      *upper_bound = node->super.key;
    }
  }
  if (SF_DEBUG) {
    putc('\n', stderr);
  }
  return queue;
}

void sf_solve(const struct sf_input *input, struct sf_greenhouse_list *solution) {
  *solution = SF_EMPTY_GREENHOUSE_LIST;
  struct sf_interval2 box;
  if (!sf_bounding_box_from_input(input, &box)) {
    return;
  }
  double upper_bound = INFINITY;
  struct sf_node *last = NULL;
  int next_id = 0;
  struct sf_node *root = allocate(&last, NULL, &next_id);
  root->super.key = 0.0;
  for (int i = box.i.min; i < box.i.max; i++) {
    for (int j = box.j.min; j < box.j.max; j++) {
      root->y[i][j] = 0.0;
    }
  }
  root->tax = 0.0;
  for (int i = box.i.min; i < box.i.max + 1; i++) {
    for (int j = box.j.min; j < box.j.max + 1; j++) {
      root->y_sum[i][j] = 0.0;
    }
  }
  root->list = sf_greenhouse_list_from_input(input);
  root->solution = SF_EMPTY_GREENHOUSE_LIST;
  struct pq_node *top = evaluate_node(input, box, NULL, root, solution, &upper_bound);
  while (top && top->key < upper_bound) {
    struct sf_node *node = (struct sf_node *)top;
    top = pq_pop(top);
    while (1) {
      struct pq_node *top_child = NULL;
      for (int k = 1; k <= 2; k++) {
        struct sf_node *child = allocate(&last, node, &next_id);
        child->super.key = node->super.key;
        for (int i = box.i.min; i < box.i.max; i++) {
          for (int j = box.j.min; j < box.j.max; j++) {
            child->y[i][j] = node->y[i][j];
          }
        }
        child->tax = node->tax;
        for (int i = box.i.min; i < box.i.max + 1; i++) {
          for (int j = box.j.min; j < box.j.max + 1; j++) {
            child->y_sum[i][j] = node->y_sum[i][j];
          }
        }
        child->exclude = k;
        child->solution = sf_copy(sf_partition(child, upper_bound, node->solution));
        top_child = evaluate_node(input, box, top_child, child, solution, &upper_bound);
      }
      if (!(top_child && top_child->key < upper_bound)) {
        break;
      }
      node = (struct sf_node *)top_child;
      top_child = pq_pop(top_child);
      if (top_child && top_child->key < upper_bound) {
        top = pq_push(top, top_child);
      }
    }
  }
  free(root->list.greenhouse);
  do {
    struct sf_node *previous = last->previous;
    free(last->solution.greenhouse);
    free(last);
    last = previous;
  } while (last);
}
